-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
M.type = "auxiliary"
M.relevantDevice = nil

local hasRegisteredQuickAccess = false

local defaultSettings = {}

local driveModes = {}
local enabledDriveModes = {}

local activeDriveMode = nil
local activeDriveModeIndex = nil

local function setExhaustGainMufflingOffset(engineName, mufflingOffset, gainOffset)
  local engine = powertrain.getDevice(engineName)
  if not engine or not engine.setExhaustGainMufflingOffset then
    return
  end

  engine:setExhaustGainMufflingOffset(mufflingOffset, gainOffset)
end

local function setTurbochargerWastegateOffset(engineName, wastegateOffset)
  local engine = powertrain.getDevice(engineName)
  if not engine or not engine.turbocharger or not engine.turbocharger.setWastegateOffset then
    return
  end

  engine.turbocharger.setWastegateOffset(wastegateOffset)
end

local function setLimit(engineName, tempLimit)
  local engine = powertrain.getDevice(engineName)
  if not engine or not engine.setLimit then
    return
  end

  engine:setLimit(tempLimit)
end

local function setGearboxMode(mode)
  if controller.mainController.setDefaultForwardMode then
    controller.mainController.setDefaultForwardMode(mode)
  end
end

local function applyDriveMode(mode)
  for _, setting in pairs(mode.settings) do
    if setting.type == "exhaust" then
      setExhaustGainMufflingOffset(setting.name, setting.mufflingOffset, setting.gainOffset)
    elseif setting.type == "turbocharger" then
      setTurbochargerWastegateOffset(setting.name, setting.wastegateOffset)
    elseif setting.type == "transmission" then
      setGearboxMode(setting.defaultForwardMode)
	elseif setting.type == "limit" then
      setLimit(setting.name, setting.tempLimit)
    elseif setting.type == "controller" then
      local controller = controller.getController(setting.controllerName)
      if controller and controller.setParameters then
        local controllerSettings = deepcopy(setting)
        controllerSettings.type = nil
        controllerSettings.key = nil
        controllerSettings.controllerName = nil
        controller.setParameters(controllerSettings)
      end
    end
  end
end

local function setDriveMode(modeKey)
  local mode = driveModes[modeKey]
  if mode then
    activeDriveMode = mode
    applyDriveMode(mode)

    guihooks.message("Drivemode: " .. mode.name, 5, "vehicle.driveMode")
  end
end

local function nextDriveMode()
  local key
  activeDriveModeIndex, key = next(enabledDriveModes, activeDriveModeIndex)
  if not key then
    activeDriveModeIndex, key = next(enabledDriveModes)
  end
  setDriveMode(key)
end

local function previousDriveMode()
  local newIndex = activeDriveModeIndex - 1
  if newIndex < 1 then
    newIndex = #enabledDriveModes
  end
  activeDriveModeIndex = newIndex
  setDriveMode(enabledDriveModes[activeDriveModeIndex])
end

local function updateGFX(dt)
end

local function reset()
end

local function resetLastStage()
  setDriveMode(activeDriveMode.key)
end

local function serialize()
  return {activeDriveModeKey = activeDriveMode.key, activeDriveModeIndex = activeDriveModeIndex}
end

local function deserialize(data)
  if data then
    if data.activeDriveModeKey and data.activeDriveModeIndex then
      activeDriveModeIndex = data.activeDriveModeIndex
      setDriveMode(data.activeDriveModeKey)
    end
  end
end

local function registerQuickAccess()
  if not hasRegisteredQuickAccess then
    core_quickAccess.addEntry(
      {
        level = "/",
        generator = function(entries)
          table.insert(
            entries,
            {
              title = "ui.radialmenu2.ESC",
              priority = 40,
              icon = "radial_regular_esc",
              onSelect = function()
                controller.getControllerSafe("driveModes").nextDriveMode()
                return {"reload"}
              end
            }
          )
        end
      }
    )
    hasRegisteredQuickAccess = true
  end
end

local function init(jbeamData)
  defaultSettings = {}
  if jbeamData.defaultSettings then
    local defaults = jbeamData.defaultSettings
    local defaultSettingData = tableFromHeaderTable(defaults)
    for _, v in pairs(defaultSettingData) do
      local key = v.key or v.controllerName or v.type
      defaultSettings[key] = v
    end
  end

  local modeData = jbeamData.modes
  if not modeData then
    return
  end

  local enabledModeKeys = {}
  for k, v in pairs(jbeamData) do
    if k:sub(1, #"enabledModes") == "enabledModes" then
      for _, modeKey in pairs(v) do
        enabledModeKeys[modeKey] = true
      end
    end
  end

  --dump(enabledModeKeys)

  local modeSorting = {}

  for k, mode in pairs(modeData) do
    if driveModes[mode.order] then
      log("E", "driveModes.init", string.format("Duplicate mode order (%d) with mode: %s", mode.order, mode.name))
    end

    driveModes[k] = {
      key = k,
      name = mode.name,
      order = mode.order,
      settings = deepcopy(defaultSettings)
    }

    local modeSettings = tableFromHeaderTable(mode.settings or {})
    if modeSettings then
      for _, newSetting in pairs(modeSettings) do
        local key = newSetting.key or newSetting.controllerName or newSetting.type
        driveModes[k].settings[key] = newSetting
      end
    end

    if enabledModeKeys[k] then
      table.insert(modeSorting, {key = k, order = mode.order})
    end
  end

  table.sort(
    modeSorting,
    function(a, b)
      local ra, rb = a.order or 0, b.order or 0
      return ra < rb
    end
  )
  --dump(modeSorting)

  --dump(driveModes)

  for _, mode in pairs(modeSorting) do
    table.insert(enabledDriveModes, mode.key)
  end

  --dump(enabledDriveModes)

  -- for name, mode in pairs(modeSorting) do
  --   if driveModeOrder[mode.order] then
  --     log("E", "driveModes.init", string.format("Duplicate mode order (%d) with mode: %s", mode.order, mode.name))
  --   end
  --   driveModeOrder[mode.order] = name
  -- end

  --dump(driveModeOrder)

  registerQuickAccess()
end

local function initLastStage(jbeamData)
  local key
  local requestedDefaultMode = jbeamData.defaultMode or ""
  if driveModes[requestedDefaultMode] then
    for k, modeName in ipairs(enabledDriveModes) do
      if modeName == requestedDefaultMode then
        key = requestedDefaultMode
        activeDriveModeIndex = k
      end
    end
  end

  if not key then
    activeDriveModeIndex, key = next(enabledDriveModes)
  end
  setDriveMode(key)
end

M.init = init
M.reset = reset
M.resetLastStage = resetLastStage
M.initLastStage = initLastStage
M.updateGFX = updateGFX
M.serialize = serialize
M.deserialize = deserialize

M.nextDriveMode = nextDriveMode
M.previousDriveMode = previousDriveMode
M.setDriveMode = setDriveMode

return M
